import json
import os
import platform
import shutil
import sys
import tempfile
import threading
from subprocess import PIPE, Popen

from demogpt.plan.controllers import checkPromptTemplates, refineKeyTypeCompatiblity
from demogpt.plan.chains.task_chains import TaskChains


def init(title=""):
    if title:
        return IMPORTS_CODE_SNIPPET + f"\nst.title({title})\n"
    return IMPORTS_CODE_SNIPPET


def getCodeSnippet(task,code_snippets,iters):
    task = refineKeyTypeCompatiblity(task)
    task_type = task["task_type"]
    code = ""
    if task_type == "ui_input_text":
        code = TaskChains.uiInputText(task=task,code_snippets=code_snippets)
    elif task_type == "ui_output_text":
        code = TaskChains.uiOutputText(task=task,code_snippets=code_snippets)
    elif task_type == "prompt_chat_template":
        res = ""
        is_valid = False
        res = TaskChains.promptChatTemplate(task=task,code_snippets=code_snippets)
        index = 0
        while not is_valid:
            check = checkPromptTemplates(res,task)
            is_valid = check["valid"]
            feedback = check["feedback"]
            if not is_valid:
                res = TaskChains.promptTemplateRefiner(res,feedback)
            else:
                break   
            index += 1
            if index == iters:
                break
        code = getPromptChatTemplateCode(res, task)
    elif task_type == "path_to_content":
        code = TaskChains.pathToContent(task=task,code_snippets=code_snippets)
    elif task_type == "doc_to_string":
        code = TaskChains.docToString(task=task,code_snippets=code_snippets)
    elif task_type == "string_to_doc":
        code = TaskChains.stringToDoc(task=task,code_snippets = code_snippets)
    elif task_type == "ui_input_file":
        code = TaskChains.uiInputFile(task=task,code_snippets=code_snippets)
    elif task_type == "doc_loader":
        code = TaskChains.docLoad(task=task,code_snippets=code_snippets)
    elif task_type == "doc_summarizer":
        code = TaskChains.summarize(task=task,code_snippets=code_snippets)
    return code.strip() + "\n"


def refine(code):
    if "```" in code:
        code = code.split("```")[1]
        if code.startswith("python"):
            code = code[len("python") :].strip()
    return code


def getPromptChatTemplateCode(templates, task):
    inputs = task["input_key"]
    variable = task["output_key"]
    run_call = "{}"

    if inputs == "none":
        signature = f"{templates['function_name']}()"
        function_call = f"{variable} = {signature}"
    else:
        if isinstance(inputs, str):
            if inputs.startswith("["):
                inputs = inputs[1:-1]
            inputs = [var.strip() for var in inputs.split(",")]
        if len(inputs) > 0:
            run_call = ", ".join([f"{var}={var}" for var in inputs])
        signature = f"{templates['function_name']}({','.join(inputs)})"
        function_call = f"""
if {' and '.join(inputs)}:
    {variable} = {signature}
else:
    {variable} = ""
"""

    temperature = 0 if templates.get("variety", "False") == "False" else 0.7

    code = f"""\n
def {signature}:
    chat = ChatOpenAI(
        model="gpt-3.5-turbo-16k",
        temperature={temperature}
    )
    system_template = \"\"\"{templates['system_template']}\"\"\"
    system_message_prompt = SystemMessagePromptTemplate.from_template(system_template)
    human_template = \"\"\"{templates['template']}\"\"\"
    human_message_prompt = HumanMessagePromptTemplate.from_template(human_template)
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )

    chain = LLMChain(llm=chat, prompt=chat_prompt)
    result = chain.run({run_call})
    return result # returns string   

{function_call}               

"""
    return code


def runThread(proc):
    proc.communicate()


def runStreamlit(code, openai_api_key):
    """
    Runs the provided code as a Streamlit application and returns the process ID.

    Args:
        code (str): The code of the Streamlit application.

    Returns:
        int: The process ID of the Streamlit application.
    """
    tmp = tempfile.NamedTemporaryFile("w", suffix=".py", delete=False, encoding="utf-8")
    tmp.write(code)
    tmp.flush()
    environmental_variables = {
        "OPENAI_API_KEY": openai_api_key,
        "STREAMLIT_SERVER_PORT": "8502",
    }
    streamlit_path = shutil.which("streamlit")
    if True or platform.system() == "Windows":
        env = os.environ.copy()
        env["PYTHONPATH"] = ""
        env["OPENAI_API_KEY"] = openai_api_key
        env["STREAMLIT_SERVER_PORT"] = "8502"
        python_path = sys.executable
        process = Popen(
            [python_path, "-m", "streamlit", "run", tmp.name],
            env=env,
            stdout=PIPE,
            stderr=PIPE,
        )
        threading.Thread(target=runThread, args=(process,)).start()
    try:
        tmp.close()
    except PermissionError:
        pass

    return process.pid


IMPORTS_CODE_SNIPPET = """
import streamlit as st
from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)
from langchain.document_loaders import *
from langchain.chains.summarize import load_summarize_chain
import tempfile
from langchain.docstore.document import Document
"""



import json
import os

from .chains import prompts
from .. import utils

from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)


class TaskChains:
    llm = None

    @classmethod
    def setLlm(
        cls, model, openai_api_key=os.getenv("OPENAI_API_KEY", ""), temperature=0
    ):
        cls.llm = ChatOpenAI(
            model=model, openai_api_key=openai_api_key, temperature=temperature
        )

    @classmethod
    def getChain(cls, system_template="", human_template="", **kwargs):
        prompts = []
        if system_template:
            prompts.append(SystemMessagePromptTemplate.from_template(system_template))
        if human_template:
            prompts.append(HumanMessagePromptTemplate.from_template(human_template))
        chat_prompt = ChatPromptTemplate.from_messages(prompts)
        return LLMChain(llm=cls.llm, prompt=chat_prompt).run(**kwargs)

    @classmethod
    def uiInputText(cls, task,code_snippets):
        variable = task["output_key"]
        instruction = task["description"]
        code = cls.getChain(
            human_template=prompts.ui_input_text.human_template,
            instruction=instruction,
            variable=variable,
            code_snippets=code_snippets
        )
        return utils.refine(code)

    @classmethod
    def uiOutputText(cls, task,code_snippets):
        args = task["input_key"]
        if isinstance(args, list):
            args = ",".join(args)
        instruction = task["description"]
        code = cls.getChain(
            human_template=prompts.ui_output_text.human_template,
            instruction=instruction,
            args=args,
            code_snippets=code_snippets
        )
        return utils.refine(code)

    @classmethod
    def uiInputFile(cls, task,code_snippets):
        variable = task["output_key"]
        instruction = task["description"]
        code = cls.getChain(
            system_template=prompts.ui_input_file.system_template,
            human_template=prompts.ui_input_file.human_template,
            instruction=instruction,
            variable=variable,
            code_snippets=code_snippets
        )
        return utils.refine(code)
    
    @classmethod
    def pathToContent(cls, task,code_snippets):
        instruction = task["description"]
        argument = task["input_key"]
        variable = task["output_key"]

        code = cls.getChain(
            system_template=prompts.path_to_file.system_template,
            human_template=prompts.path_to_file.human_template,
            instruction=instruction,
            argument=argument,
            variable=variable,
            code_snippets=code_snippets
        )
        return utils.refine(code)

    @classmethod
    def promptChatTemplate(cls, task,code_snippets):
        inputs = task["input_key"]
        instruction = task["description"]

        res = cls.getChain(
            system_template=prompts.prompt_chat_template.system_template,
            human_template=prompts.prompt_chat_template.human_template,
            instruction=instruction,
            inputs=inputs,
            code_snippets=code_snippets
        )
        res = res[res.find("{"):res.rfind("}")+1]
        return json.loads(res)
    
    @classmethod
    def promptTemplateRefiner(cls, templates,feedback):
        res = cls.getChain(
            system_template=prompts.prompt_chat_refiner.system_template,
            human_template=prompts.prompt_chat_refiner.human_template,
            templates=templates,
            feedback=feedback
        )
        res = res[res.find("{"):res.rfind("}")+1]
        return json.loads(res)
    
    @classmethod
    def docLoad(cls, task,code_snippets):
        instruction = task["description"]
        argument = task["input_key"]
        if isinstance(argument,list):
            argument = argument[0]
        variable = task["output_key"]
        function_name = task["task_name"]

        code = cls.getChain(
            system_template=prompts.doc_load.system_template,
            human_template=prompts.doc_load.human_template,
            instruction=instruction,
            argument=argument,
            variable=variable,
            function_name=function_name,
            code_snippets=code_snippets
        )
        return utils.refine(code)
    
    @classmethod
    def stringToDoc(cls, task, code_snippets):
        argument = task["input_key"]
        variable = task["output_key"]
        code =f'{variable} = "\".join([doc.page_content for doc in {argument}])'
        code = f"""
from langchain.docstore.document import Document
{variable} =  [Document(page_content={argument}, metadata={{'source': 'local'}})]
        """
        return code
    
    @classmethod
    def docToString(cls, task, code_snippets):
        argument = task["input_key"]
        variable = task["output_key"]
        code =f'{variable} = "\".join([doc.page_content for doc in {argument}])'
        return code
    
    @classmethod
    def summarize(cls, task,code_snippets):
        argument = task["input_key"]
        variable = task["output_key"]
        function_name = task["task_name"]

        code = f"""
from langchain.chat_models import ChatOpenAI
from langchain.chains.summarize import load_summarize_chain
def {function_name}(docs):
    llm = ChatOpenAI(temperature=0, model_name="gpt-3.5-turbo-16k")
    chain = load_summarize_chain(llm, chain_type="stuff")
    return chain.run(docs)
if {argument}:
    {variable} = summarize(argument)
else:
    variable = ""
"""
        return code




import os
import sys

sys.path.append(os.path.abspath('src/plan/'))
import json
import unittest

import utils
from chains.chains import Chains
from chains.task_chains import TaskChains
from test_cases import TEST_CASES, TOOL_EXAMPLES, INSTRUCTIONS, CODE_SNIPPETS
from tqdm import tqdm


class TestDemoGPT(unittest.TestCase):
    TEST_INDEX = 6
    INSTRUCTION = INSTRUCTIONS[TEST_INDEX]
    REFINE_ITERATIONS = 10
    #"Create a system that can summarize a content taken from url then create a blog post on the summarization"
    #"Create a system that can solve any math problem"
    TITLE = "My App"

    @classmethod
    def setUpClass(cls):
        cls.f = open(f"test_{TestDemoGPT.TEST_INDEX}.log", "w")
        
        # it sets the model name
        model_name = "gpt-3.5-turbo-0613"

        Chains.setLlm(model_name)
        TaskChains.setLlm(model_name)
        
    @classmethod
    def writeToFile(cls,title,res,instruction):
        cls.f.write(title)
        cls.f.write("\n")
        cls.f.write(instruction)
        cls.f.write("\n")
        cls.f.write(res)
        cls.f.write("\n")
        cls.f.flush()

    @classmethod
    def writeFinalToFile(cls,res,instruction):
        with open("test_final.py","w") as f:
            f.write("#"+instruction+"\n")
            f.write(res)
            f.flush()

    def test_plan(self):
        for test_case in tqdm(TEST_CASES):
            instruction = test_case["instruction"]
            instruction = "Create a system that can summarize a website from the given URL."
            plan = Chains.plan(instruction)
            self.writeToFile("PLAN",plan,instruction)
            break

    def test_tasks(self):
        for test_case in tqdm(TEST_CASES):
            instruction = test_case["instruction"]
            plan = test_case["plan"]
            task_list = Chains.tasks(instruction=instruction,plan=plan)
            self.writeToFile("TASK LIST",json.dumps(task_list, indent=4),instruction)

    def test_final(self):
        for test_case in tqdm(CODE_SNIPPETS):
            instruction = test_case["instruction"]
            plan = test_case["plan"]
            code_snippets = test_case["code_snippets"]
            code_snippets = utils.IMPORTS_CODE_SNIPPET + code_snippets
            final_code = Chains.final(instruction=instruction,
                                      code_snippets=code_snippets,
                                      plan=plan
                                      )
            self.writeFinalToFile(final_code,instruction)     
            break       
    
    def test_task_ui_input_text(self):
        for example in TOOL_EXAMPLES["ui_input_text"]:
            instruction = example["instruction"]
            variable = example["variable"]
            res = TaskChains.uiInputText(instruction=instruction,variable=variable)
            self.writeToFile("UI INPUT TEXT",res,instruction)

    def test_task_ui_output_text(self):
        for example in TOOL_EXAMPLES["ui_input_text"]:
            instruction = example["instruction"]
            args = example["args"]
            res = TaskChains.uiOutputText(instruction=instruction,args=args)
            self.writeToFile("UI OUTPUT TEXT",res,instruction)

    def test_task_prompt_chat_template(self):
        for example in TOOL_EXAMPLES["ui_input_text"]:
            instruction = example["instruction"]
            inputs = example["inputs"]
            res = TaskChains.promptChatTemplate(instruction=instruction,inputs=inputs)
            self.writeToFile("PROMPT CHAT TEMPLATE",res,instruction)

    def test(self):
        title = TestDemoGPT.TITLE

        instruction = TestDemoGPT.INSTRUCTION

        plan = Chains.plan(instruction)

        self.writeToFile("PLAN",plan,instruction)

        task_list = Chains.tasks(instruction=instruction, plan=plan)

        self.writeToFile("TASK LIST",json.dumps(task_list, indent=4),instruction)
        
        task_controller_result = Chains.taskController(tasks=task_list)
        
        self.writeToFile("TASK CONTROLLER RESULT",json.dumps(task_controller_result, indent=4),instruction)
        for _ in range(TestDemoGPT.REFINE_ITERATIONS):
            if not task_controller_result["valid"]:
                task_list = Chains.refineTasks(instruction=instruction, tasks=task_list, feedback = task_controller_result["feedback"])
                self.writeToFile("REFINED TASK LIST",json.dumps(task_list, indent=4),instruction)
                task_controller_result = Chains.taskController(tasks=task_list)
            else:
                break
        
            self.writeToFile("FEEDBACK",task_controller_result["feedback"],instruction)
            
        
    
        code_snippets = utils.init(title)

        self.writeToFile("CODE SNIPPETS","",instruction)

        for task in tqdm(task_list):
            code = utils.getCodeSnippet(task,code_snippets)
            code = "#"+task["description"] + "\n" + code
            code_snippets += code
            self.writeToFile("",code,"")

        draft_code = Chains.draft(instruction=instruction,
                                  code_snippets=code_snippets,
                                  plan=plan
                                  )
        
        self.writeToFile("COMBINED CODE",draft_code,instruction)
        
        final_code = Chains.final(draft_code=draft_code)
        
        self.writeFinalToFile(final_code,instruction)

    def test_all(self):
        for test_case in tqdm(TEST_CASES):
            instruction = test_case["instruction"]
            TestDemoGPT.test(instruction)
        TestDemoGPT.f.close()
        
    
if __name__ == '__main__':
    TestDemoGPT.INSTRUCTION = os.environ.get('instruction', TestDemoGPT.INSTRUCTION)
    TestDemoGPT.TITLE = os.environ.get('title', TestDemoGPT.TITLE)
    unittest.test()


from model import Model


def create(idea, title):
    # Move the logic you want to be available as a library here
    # This function should be a generator that yields results based on the given idea and title

    """for data in generate_response(idea, title):  # assuming generate_response is the function you want to use
        yield data"""
    ...


human_template = """
Write and call a function which shows text by streamlit text element code depending on the instruction below:
Suppose that, streamlit has been imported by "import streamlit as st" so you don't need to import it.

The function arguments are "{args}".
You can assume that the variables with the same names are already defined.
So you don't need to give dummy variables to the function while calling.
You should use "{args}" to call the function
In the function, also add a descriptive part next to the "{args}"
After defining the function, call it with "{args}". 
Assume that, they have been already defined but only call the function by checking if the input is not an empty string.
use st.markdown to show the main text

Here is the part of the code that you are supposed to continue:
{code_snippets}

Instruction:{instruction}
Streamlit Code:
"""



import json
import os

from . import prompts

from .. import utils
from ..controllers import checkDTypes

from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)

class Chains:
    @classmethod
    def setLlm(
        cls, model, openai_api_key=os.getenv("OPENAI_API_KEY", ""), temperature=0
    ):
        cls.llm = ChatOpenAI(
            model=model, openai_api_key=openai_api_key, temperature=temperature
        )

    @classmethod
    def getChain(cls, system_template="", human_template="", **kwargs):
        prompts = []
        if system_template:
            prompts.append(SystemMessagePromptTemplate.from_template(system_template))
        if human_template:
            prompts.append(HumanMessagePromptTemplate.from_template(human_template))
        chat_prompt = ChatPromptTemplate.from_messages(prompts)
        return LLMChain(llm=cls.llm, prompt=chat_prompt).run(**kwargs)

    @classmethod
    def plan(cls, instruction):
        return cls.getChain(
            system_template=prompts.plan.system_template,
            human_template=prompts.plan.human_template,
            instruction=instruction,
        )

    @classmethod
    def tasks(cls, instruction, plan):
        task_list = cls.getChain(
            system_template=prompts.tasks.system_template,
            human_template=prompts.tasks.human_template,
            instruction=instruction,
            plan=plan
        )
        return json.loads(task_list)
    
    @classmethod
    def taskController(cls, tasks):
        return checkDTypes(tasks)
    
    @classmethod
    def refineTasks(cls, instruction, tasks, feedback):
        task_list = cls.getChain(
            system_template=prompts.task_refiner.system_template,
            human_template=prompts.task_refiner.human_template,
            instruction=instruction,
            tasks=tasks,
            feedback=feedback
        )
                
        return json.loads(task_list)
    
    @classmethod
    def draft(cls, instruction, code_snippets, plan):
        code = cls.getChain(
            system_template=prompts.combine.system_template,
            human_template=prompts.combine.human_template,
            instruction=instruction,
            code_snippets=code_snippets,
            plan=plan,
        )
        return utils.refine(code)

    @classmethod
    def final(cls, draft_code):
        code = cls.getChain(
            system_template=prompts.final.system_template,
            human_template=prompts.final.human_template,
            draft_code=draft_code
        )
        return utils.refine(code)



system_template = """
You are a coding assistant specialized in working with Streamlit applications and error-free code generation. 
Your task is, paying special attention to user input handling, state management and not getting any "not defined" error because of if statements without else. 
Don't forget to add title with already defined st.title line
Generate nothing else but only the code so that it can be directly used.
"""

human_template = """
Refine the Original Code like in the following order:

Step-1 Write all the import statements from the Draft Code (only the ones that are used).

Step-2 Write all the function definitions

Step-3 Get input from the user

Step-4 Put a submit button with an appropriate title

Step-5 Call functions only if all user inputs are taken and the button is clicked.


Draft Code: 
{draft_code}
################################
Final Error-free Code:
"""


loaders = """
For Local TXT file:
from langchain.document_loaders import TextLoader
loader = TextLoader(<local_txt_file_path>)
################################
For Web Page:
from langchain.document_loaders import WebBaseLoader
loader = WebBaseLoader("<url>")
################################
For Online PDF:
from langchain.document_loaders import OnlinePDFLoader
loader = OnlinePDFLoader("<online_pdf_url>")
################################
For Local PDF:
from langchain.document_loaders import UnstructuredPDFLoader
loader = UnstructuredPDFLoader(
    <local_pdf_full_path>, mode="elements", strategy="fast"
    )
################################
For Power Point:
from langchain.document_loaders import UnstructuredPowerPointLoader
loader = UnstructuredPowerPointLoader(
    <local_powerpoint_file>, mode="elements", strategy="fast"
    )
################################
For CSV:
from langchain.document_loaders.csv_loader import UnstructuredCSVLoader
loader = UnstructuredCSVLoader(<csv_file_path>, mode="elements")
################################
For Excel:
from langchain.document_loaders.excel import UnstructuredExcelLoader
loader = UnstructuredExcelLoader(<excel_file_path>, mode="elements")
"""

system_template = f"""
These are the Loader classes that you should select.
Select the loader according to the input type unless the input type is ambiguous.
{loaders}
"""

human_template = """
Write a loader function using langchain.document_loaders 
to load the document for the argument name, variable and instruction 
below like in the below format:

###
def {function_name}(argument):
    loader = Loader(path) # Select the appropriate Loader
    docs = loader.load()
    return docs

if {argument}:
    {variable} = {function_name}({argument})
###
    
While using the loader, don't change "mode" and "strategy" arguments, they need to be constant as stated.
If there are no such arguments, ignore it.

Instruction:{instruction}    

Document Loader Code:
"""


from .task_definitions import TASK_DESCRIPTIONS

system_template = f"""
Refine the Generated Task List according to the Feedback

Task objects must be Python dictionaries, and the output should strictly conform to a Python list of JSON objects.
So use double quotes because this output will be converted to json with json.loads function.

It is extremely important to include these keys in each task object:

-step
-task_type
-task_name
-input_key
-input_data_type
-output_key
-output_data_type
-description

################################

Available Tasks:

{TASK_DESCRIPTIONS}
"""

human_template = """
Instruction:{instruction}
################################
Generated Task List:
{tasks}
################################
Feedback given to the Generated Task (extremly important) : {feedback}
################################
New Refined Task List:
"""


human_template = """
Write a streamlit textfield or textarea code depending on the instruction below:
Suppose that, streamlit has been imported by "import streamlit as st" so you don't need to import it.
Here is the part of the code that you are supposed to continue:
{code_snippets}

Assign the taken input to the variable called "{variable}"

Instruction:{instruction}
Streamlit Code:
"""



from demogpt.plan.chains.prompts.task_definitions import TASK_TYPE2_TASK

def refineKeyTypeCompatiblity(task):
    if task["input_data_type"] == "none":
        if task["input_key"] != "none":
            task["input_key"] = "none"
    if task["output_data_type"] == "none":
        if task["output_key"] != "none":
            task["output_key"] = "none"
    return task
            
def checkDTypes(tasks):
    
    feedback = ""
    for task in tasks:
        name = task["task_type"]

        input_data_type = task["input_data_type"]
        output_data_type = task["output_data_type"]
        input_key = task["input_key"]
        
        reference = TASK_TYPE2_TASK[name]
        reference_input = reference["input_data_type"]
        reference_output = reference["output_data_type"]
        
        if task["step"] == 1:
            if input_key != "none":
                feedback += f"Since {name} is the first task, its input data type is supposed to be none but it is {input_key}.Please find another way.\n"
        
        # Check input data types
        elif reference_input.startswith("*"):
            reference_input = reference_input.replace("*", "")
            if isinstance(input_key,str):
                if input_data_type != reference_input and input_data_type != "none":
                    feedback += f"""
                    {name} expects all inputs as {reference_input} or none but the data type of {input_key} is {input_data_type} not {reference_input}. Please find another way.\n
                    """
            else: 
                for res,data_type in zip(input_key,input_data_type):
                    if data_type != reference_input:
                        feedback += f"""
                        {name} expects all inputs as {reference_input} but data type of {res} is {data_type} not {reference_input}. Please find another way.\n
                        """
        elif input_data_type != reference_input:
            feedback += f"""
            {name} expects all inputs as {reference_input} but the data type of {input_key} is {input_data_type} not {reference_input}. Please find another way.\n
            """
            
        # Check output data types
        if output_data_type != reference_output:
            feedback += f"""
            {name} should output in {reference_output} data type but it is {output_data_type} not {reference_output}. Please find another way.\n
            """
        
    valid = len(feedback) == 0
    
    return {
        "feedback": feedback,
        "valid": valid
    }
    
def checkPromptTemplates(templates,task):
    human_template = templates["template"]
    system_template = templates["system_template"]
    templates = human_template + system_template
    inputs = task["input_key"]
    if inputs == "none":
        inputs = []
    else:
        if isinstance(inputs, str):
            if inputs.startswith("["):
                inputs = inputs[1:-1]
            inputs = [var.strip() for var in inputs.split(",")]
    feedback = ""
    for input_key in inputs:
        if f"{{{input_key}}}" not in templates:
            feedback += f"{{{input_key}}} is not included in any of the templates. Please add it.\n"
    
    valid = len(feedback) == 0
    
    return {
        "feedback":feedback,
        "valid":valid
    }
        
                
        
    
    


import json
import re

AVAILABLE_TASKS_COUNT = 9

################################

ALL_TASKS = [
    {
        "name": "ui_input_text",
        "description": "Gets input from the user via a text field.",
        "good_at": "Retrieving text input from the user.",
        "input_data_type": "none",
        "output_data_type": "string",
        "purpose": "Collect user-entered text for further processing.",
    },
    {
        "name": "ui_input_file",
        "description": "Provide a mechanism for users to upload a file and return its path. The task involves creating a file upload widget and returning its file path",
        "good_at": "Enabling file uploads and making the file path available for doc_load",
        "input_data_type": "none",
        "output_data_type": "string",
        "purpose": "Getting local file path with upload file widget so that doc_load can use this path",
    },
    {
        "name": "ui_output_text",
        "description": "Shows text output to the user.",
        "good_at": "Showing text to the user.",
        "input_data_type": "string",
        "output_data_type": "none",
        "purpose": "Displaying textual information to the user.",
    },
    {
        "name": "prompt_chat_template",
        "description": "Generate any string output according to the given instruction by AI",
        "good_at": "Creating context-aware string, responses, role play, instructions, that can be generated by AI",
        "input_data_type": "*string",
        "output_data_type": "string",
        "purpose":"Using AI to generate smart text output from given context or instruction",
    },
    {
        "name": "doc_loader",
        "description": "Load file content from path (txt or url or pdf file path or csv file path or powerpoint file path) and generate docs",
        "good_at": "Loading from external sources only from url or path not the content",
        "input_data_type":"string",
        "output_data_type":"Document",
        "purpose": "Loading external files",
    },
    {
        "name": "doc_to_string",
        "description": "Convert Document object to string",
        "good_at": "Converting Document object to string where the next task is expecting string instead of Document object",
        "input_data_type":"Document",
        "output_data_type":"string",
        "purpose": "Converting Document object to string",
    },
    {
        "name": "string_to_doc",
        "description": "Convert string to Document object",
        "good_at": "Converting string to Document object where the next task is expecting Document object instead of string",
        "input_data_type":"string",
        "output_data_type":"Document",
        "purpose": "Converting string to Document object",
    },
    {
        "name": "doc_summarizer",
        "description": "Summarize Document Objects",
        "good_at": "Summarizing long Document Objects",
        "input_data_type":"Document",
        "output_data_type":"string",
        "purpose": "Summarization of long Document Objects",
    },
    {
        "name": "path_to_content",
        "description": "Retrieve the content of the file from given path",
        "good_at": "Retrieving the content of the file from given path",
        "input_data_type":"string",
        "output_data_type":"string",
        "purpose": "Retrieving the content of the file from given path",
    },
    {
        "name": "hub_question_answering",
        "description": "Answer questions related to the file.",
        "good_at": "Question answering on files",
        "input_data_type":["string", "file"],
        "output_data_type":"string",
        "purpose": "Extracting and providing specific information from files in response to questions.",
    },
    {
        "name": "memory",
        "description": "Returns memory that could be attached as an input to any prompt_chat_template.",
        "good_at": "Memorizing the conversation history or context.",
        "input_data_type":"none",
        "output_data_type":"Memory",
        "purpose": "Storing and retrieving conversation history or contextual information.",
    },
    {
        "name": "prompt_list_parser",
        "description": "Transform the input text into a list.",
        "good_at": "Transforming text into a list.",
        "input_data_type":"string",
        "output_data_type":"list",
        "purpose": "Converting textual data into structured list format.",
    },
    {
        "name": "router",
        "description": "When there are multiple prompt_chat_template objects, it uses the appropriate one to answer the question.",
        "good_at": "Handling different types of questions that require different abilities.",
        "input_data_type":"*prompt_chat_template",
        "output_data_type":"string",
        "purpose": "Routing queries to the appropriate handler based on context or type.",
    },
    {
        "name": "react",
        "description": "Answer questions that require external search on the web.",
        "good_at": "Answering questions that require Google search or other web searches.",
        "input_data_type":"string",
        "output_data_type":"string",
        "purpose": "Finding information online to answer user queries.",
    },
    {
        "name": "cpal_chain",
        "description": "Solve math problems end to end",
        "good_at": "Directly solving any math problems",
        "input_data_type":"string",
        "output_data_type":"string",
        "purpose": "Performing mathematical calculations and solving problems based on the input question",
    },
    {
        "name": "hub_bash",
        "description": "Do operations on the bash by running needed scripts on the terminal to apply the command.",
        "good_at": "Executing bash commands and providing results.",
        "input_data_type":"string",
        "output_data_type":"string",
        "purpose": "Running scripts or commands on the terminal and returning the output.",
    },
    {
        "name": "hub_meteo",
        "description": "Gives weather-related information from the question.",
        "good_at": "Answering weather-related questions.",
        "input_data_type":"string",
        "output_data_type":"string",
        "purpose": "Providing weather forecasts, conditions, and related information.",
    }
]

def jsonFixer(data):
    data = json.dumps(data, indent=4)
    return data.replace("{", "{{").replace("}", "}}")

TASKS = ALL_TASKS[:AVAILABLE_TASKS_COUNT]

TASK_NAMES = [task["name"] for task in TASKS]

TASK_DESCRIPTIONS = jsonFixer(TASKS)

TASK_DTYPES = {
    task["name"]:{
        "input_data_type":task["input_data_type"],
        "output_data_type":task["output_data_type"]
        }
    for task in TASKS
}

TASK_DTYPES = jsonFixer(TASK_DTYPES)

TASK_TYPE2_TASK = {
    task["name"]:task
    for task in TASKS
}


system_template = """
Write a streamlit file uploader code and return the path of uploaded file as a string depending on the instruction given to you:
Since st.file_uploader.name does not give full file path, you first need to save it then get a full file path like in the following:

uploaded_file = st.file_uploader("Upload Txt File", type=["txt"])
if uploaded_file is not None:
    # Create a temporary file to store the uploaded content
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(uploaded_file.read())
        {variable} = temp_file.name # it shows the file path

Suppose that, streamlit has been imported by "import streamlit as st" so you don't need to import it.
Here is the part of the code that you are supposed to continue:
{code_snippets}

Assign the file path to the variable called "{variable}"
You will basically use file_uploader and get file path from it but nothing else.
Not to loose the file path, please use st.session_state[<key>]
Don't read the file, only get the file path
"""

human_template = """
Instruction:{instruction}
Streamlit Code:
"""



from . import (combine, final, plan, tasks,task_controller, task_refiner, prompt_chat_refiner, task_definitions)
from .task_list import (prompt_chat_template, ui_input_file, path_to_file,
                        ui_input_text,ui_output_text, doc_load, 
                        doc_to_string, summarize)


import os
from time import sleep

from demogpt.plan.utils import init, getCodeSnippet
from demogpt.plan.chains.chains import Chains
from demogpt.plan.chains.task_chains import TaskChains
from tqdm import trange


class DemoGPT:
    def __init__(self, openai_api_key=os.getenv("OPENAI_API_KEY",""), model_name="gpt-3.5-turbo-0613",max_steps=10):
        assert openai_api_key.startswith("sk-"), "Either give openai_api_key as an argument or put it in the environment variable"
        self.model_name = model_name
        self.openai_api_key = openai_api_key
        self.max_steps = max_steps # max iteration for refining the model purpose
        Chains.setLlm(self.model_name, self.openai_api_key)
        TaskChains.setLlm(self.model_name, self.openai_api_key)

    def setModel(self, model_name):
        self.model_name = model_name
        Chains.setLlm(self.model_name, self.openai_api_key)
        TaskChains.setLlm(self.model_name, self.openai_api_key)
        
    def __repr__(self) -> str:
        return f"DemoGPT(model_name='{self.model_name}',max_steps={self.max_steps})"

    def __call__(
        self,
        instruction="Create a translation system that converts English to French",
        title="",
    ):
        yield {
            "stage": "plan",
            "completed": False,
            "percentage": 0,
            "done": False,
            "message": "Plan creation has started...",
        }

        plan = Chains.plan(instruction)

        yield {
            "stage": "plan",
            "completed": True,
            "percentage": 20,
            "done": False,
            "message": "Plan has been generated.",
        }

        sleep(1)

        yield {
            "stage": "task",
            "completed": False,
            "percentage": 30,
            "done": False,
            "message": "Task generation has started...",
        }

        task_list = Chains.tasks(instruction=instruction, plan=plan)

        yield {
            "stage": "task",
            "completed": True,
            "percentage": 50,
            "done": False,
            "message": "Tasks have been generated.",
            "tasks": task_list,
        }
        
        sleep(1)
        
        yield {
            "stage": "task",
            "completed": True,
            "percentage": 55,
            "done": False,
            "message": "Tasks are being controlled."
        }
        
        task_controller_result = Chains.taskController(tasks=task_list)
        
        for _ in trange(self.max_steps):
            if not task_controller_result["valid"]:
                task_list = Chains.refineTasks(instruction=instruction, tasks=task_list, feedback = task_controller_result["feedback"])
                task_controller_result = Chains.taskController(tasks=task_list)
            else:
                break
        
        code_snippets = init(title)

        sleep(1)

        yield {
            "stage": "draft",
            "completed": False,
            "percentage": 60,
            "done": False,
            "message": "Converting tasks to code snippets...",
        }

        num_of_tasks = len(task_list)

        for i, task in enumerate(task_list):
            code = getCodeSnippet(task,code_snippets,self.max_steps)
            code = "#"+task["description"] + "\n" + code
            code_snippets += code
            yield {
                "stage": "draft",
                "completed": i + 1 == num_of_tasks,
                "percentage": 60 + int(20 * (i + 1) / num_of_tasks),
                "done": False,
                "message": f"{i+1}/{num_of_tasks} tasks have been converted to code",
                "code": code,
            }

        sleep(1)
        
        yield {
            "stage": "draft",
            "completed": False,
            "percentage": 85,
            "done": False,
            "message": "Code snippets are being combined...",
        }
        
        draft_code = Chains.draft(instruction=instruction,
                                  code_snippets=code_snippets,
                                  plan=plan
                                  )
        
        yield {
            "stage": "draft",
            "completed": True,
            "percentage": 90,
            "done": False,
            "message": "Code snippets combined. Now code is being finalized...",
        }

        final_code = Chains.final(draft_code=draft_code)

        yield {
            "stage": "final",
            "completed": True,
            "percentage": 100,
            "done": True,
            "message": "Final code has been generated. Directing to the demo page...",
            "code": final_code,
        }
